<?php

declare(strict_types=1);

/*
 * Copyright (c) 2017-2022 François Kooman <fkooman@tuxed.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

namespace fkooman\OAuth\Server;

use DateInterval;
use DateTimeImmutable;
use fkooman\OAuth\Server\Exception\InvalidClientException;
use fkooman\OAuth\Server\Exception\InvalidGrantException;
use fkooman\OAuth\Server\Exception\InvalidRequestException;
use fkooman\OAuth\Server\Exception\SignerException;
use fkooman\OAuth\Server\Http\JsonResponse;
use fkooman\OAuth\Server\Http\RedirectResponse;
use fkooman\OAuth\Server\Http\Request;
use RuntimeException;

class OAuthServer
{
    protected const RANDOM_LENGTH = 32;

    protected DateTimeImmutable $dateTime;

    protected DateInterval $authorizationCodeExpiry;

    protected DateInterval $accessTokenExpiry;

    protected DateInterval $refreshTokenExpiry;

    private StorageInterface $storage;

    private ClientDbInterface $clientDb;

    private SignerInterface $signer;

    private ?string $issuerIdentity = null;

    public function __construct(StorageInterface $storage, ClientDbInterface $clientDb, SignerInterface $signer)
    {
        // @see https://www.php.net/manual/en/mbstring.configuration.php#ini.mbstring.func-overload
        if (false !== (bool) ini_get('mbstring.func_overload')) {
            throw new RuntimeException('"mbstring.func_overload" MUST NOT be enabled');
        }
        $this->storage = $storage;
        $this->clientDb = $clientDb;
        $this->signer = $signer;
        $this->dateTime = Dt::get();
        $this->authorizationCodeExpiry = new DateInterval('PT5M'); // 5 minutes
        $this->accessTokenExpiry = new DateInterval('PT1H');       // 1 hour
        $this->refreshTokenExpiry = new DateInterval('P90D');      // 90 days
    }

    /**
     * Set RFC 9207 Issuer Identity.
     */
    public function setIssuerIdentity(string $issuerIdentity): void
    {
        $this->issuerIdentity = $issuerIdentity;
    }

    /**
     * Validates the authorization request from the client and returns verified
     * data to show an authorization dialog.
     *
     * @return array{client_id:string,display_name:string,scope:string,redirect_uri:string}
     */
    public function getAuthorize(?Request $request = null): array
    {
        $request ??= Request::fromServerVariables();
        $clientInfo = $this->validateAuthorizeRequest($request);

        return [
            'client_id' => $request->query()->clientId(),
            'display_name' => $clientInfo->displayName(),
            'scope' => $request->query()->scope(),
            'redirect_uri' => $request->query()->redirectUri(),
        ];
    }

    /**
     * Validates the authorization request from the client and returns
     * authorize response in case the client does NOT require approval by the
     * user (resource owner).
     *
     * @return false|Http\Response
     */
    public function getAuthorizeResponse(string $userId, ?Request $request = null)
    {
        $request ??= Request::fromServerVariables();
        $clientInfo = $this->validateAuthorizeRequest($request);
        if ($clientInfo->requiresApproval()) {
            return false;
        }

        return $this->postAuthorizeApproved($userId, $request);
    }

    public function postAuthorize(string $userId, ?Request $request = null): RedirectResponse
    {
        $request ??= Request::fromServerVariables();
        $this->validateAuthorizeRequest($request);
        if ('yes' === $request->post()->approve()) {
            return $this->postAuthorizeApproved($userId, $request);
        }

        // user did not approve, tell OAuth client
        return new RedirectResponse(
            $this->prepareRedirectUri(
                $request->query()->redirectUri(),
                [
                    'error' => 'access_denied',
                    'state' => $request->query()->state(),
                ]
            )
        );
    }

    /**
     * Handles POST request to the "/token" endpoint of the OAuth server.
     */
    public function postToken(?Request $request = null): JsonResponse
    {
        $request ??= Request::fromServerVariables();

        switch ($request->post()->grantType()) {
            case 'authorization_code':
                return $this->postTokenAuthorizationCode($request);

            case 'refresh_token':
                return $this->postTokenRefreshToken($request);

            default:
                throw new InvalidRequestException('invalid "grant_type"');
        }
    }

    protected function randomBytes(): string
    {
        return random_bytes(self::RANDOM_LENGTH);
    }

    /**
     * Handles POST request to the "/authorize" endpoint of the OAuth server.
     *
     * This is typically the "form submit" on the "authorize dialog" shown in
     * the browser that the user then accepts or rejects.
     */
    private function postAuthorizeApproved(string $userId, Request $request): RedirectResponse
    {
        // every "authorization" has a unique key that is bound to the
        // authorization code, access tokens(s) and refresh token
        $authKey = Base64UrlSafe::encodeUnpadded($this->randomBytes());

        $authorizationCode = new AuthorizationCode(
            Base64UrlSafe::encodeUnpadded($this->randomBytes()),
            $authKey,
            $userId,
            $request->query()->clientId(),
            new Scope($request->query()->scope()),
            $this->dateTime->add($this->authorizationCodeExpiry),
            $request->query()->redirectUri(),
            $request->query()->codeChallenge()
        );

        return new RedirectResponse(
            $this->prepareRedirectUri(
                $request->query()->redirectUri(),
                [
                    'code' => $this->signer->sign($authorizationCode->json()),
                    'state' => $request->query()->state(),
                ]
            )
        );
    }

    /**
     * Validate the request to the "/authorize" endpoint.
     */
    private function validateAuthorizeRequest(Request $request): ClientInfo
    {
        $clientInfo = $this->getClient($request->query()->clientId());

        // make sure the provided redirect URI is supported by the client
        if (!$clientInfo->isValidRedirectUri($request->query()->redirectUri())) {
            throw new InvalidClientException('client does not support this "redirect_uri"');
        }

        return $clientInfo;
    }

    private function postTokenAuthorizationCode(Request $request): JsonResponse
    {
        $clientInfo = $this->getClient($request->post()->clientId());
        $this->verifyClientCredentials($request->post()->clientId(), $clientInfo, $request);

        try {
            // verify the authorization code signature
            $authorizationCode = AuthorizationCode::fromJson($this->signer->verify($request->post()->code()));

            // check authorization_code expiry
            if ($this->dateTime >= $authorizationCode->expiresAt()) {
                throw new InvalidGrantException('"authorization_code" expired');
            }

            // parameters in POST body need to match the parameters stored with
            // the code
            $this->verifyAuthorizationCode($request, $authorizationCode);

            // verify code_verifier
            $this->verifyCodeVerifier($request, $authorizationCode);

            // check whether authorization_code was already used
            if (null !== $this->storage->getAuthorization($authorizationCode->authKey())) {
                $this->storage->deleteAuthorization($authorizationCode->authKey());

                throw new InvalidGrantException('"authorization_code" was used before');
            }

            // delete user's expired authorizations
            $this->storage->deleteExpiredAuthorizations($authorizationCode->userId(), $this->dateTime);

            $refreshTokenExpiresAt = $this->dateTime->add($this->refreshTokenExpiry);

            // store the authorization
            $this->storage->storeAuthorization(
                $authorizationCode->userId(),
                $authorizationCode->clientId(),
                $authorizationCode->scope(),
                $authorizationCode->authKey(),
                $this->dateTime,
                $refreshTokenExpiresAt
            );

            // make sure an access_token never outlives its authorization, i.e.
            // the refresh_token expiry
            $accessTokenExpiresAt = $this->dateTime->add($this->accessTokenExpiry);
            if ($accessTokenExpiresAt > $refreshTokenExpiresAt) {
                $accessTokenExpiresAt = $refreshTokenExpiresAt;
            }

            $accessToken = new AccessToken(
                Base64UrlSafe::encodeUnpadded($this->randomBytes()),
                $authorizationCode->authKey(),
                $authorizationCode->userId(),
                $request->post()->clientId(),
                $authorizationCode->scope(),
                $accessTokenExpiresAt,
                $refreshTokenExpiresAt
            );

            $refreshToken = new RefreshToken(
                Base64UrlSafe::encodeUnpadded($this->randomBytes()),
                $authorizationCode->authKey(),
                $authorizationCode->userId(),
                $request->post()->clientId(),
                $authorizationCode->scope(),
                $refreshTokenExpiresAt
            );

            return new JsonResponse(
                [
                    'access_token' => $this->signer->sign($accessToken->json()),
                    'refresh_token' => $this->signer->sign($refreshToken->json()),
                    'token_type' => 'bearer',
                    'expires_in' => $this->expiresAtToInt($accessTokenExpiresAt),
                ],
                // The authorization server MUST include the HTTP "Cache-Control"
                // response header field [RFC2616] with a value of "no-store" in any
                // response containing tokens, credentials, or other sensitive
                // information, as well as the "Pragma" response header field [RFC2616]
                // with a value of "no-cache".
                [
                    'Cache-Control' => 'no-store',
                    'Pragma' => 'no-cache',
                ]
            );
        } catch (SignerException $e) {
            throw new InvalidGrantException(sprintf('invalid "authorization_code": %s', $e->getMessage()));
        }
    }

    private function postTokenRefreshToken(Request $request): JsonResponse
    {
        try {
            // verify the refresh code signature
            $refreshToken = RefreshToken::fromJson($this->signer->verify($request->post()->refreshToken()));

            // check refresh_token expiry
            if ($this->dateTime >= $refreshToken->expiresAt()) {
                throw new InvalidGrantException('"refresh_token" expired');
            }

            $clientInfo = $this->getClient($refreshToken->clientId());
            $this->verifyClientCredentials($refreshToken->clientId(), $clientInfo, $request);

            // scope in POST body MUST match scope stored in the refresh_token when
            // provided
            if (null !== $postScope = $request->post()->scope()) {
                if ($postScope !== (string) $refreshToken->scope()) {
                    throw new InvalidRequestException('unexpected "scope"');
                }
            }

            // make sure the authorization still exists
            if (null === $authorizationInfo = $this->storage->getAuthorization($refreshToken->authKey())) {
                throw new InvalidGrantException('"refresh_token" is no longer authorized');
            }

            // check for refresh_token replay
            if ($this->storage->isRefreshTokenReplay($refreshToken->tokenId())) {
                // it was used before, delete the complete authorization
                $this->storage->deleteAuthorization($refreshToken->authKey());

                throw new InvalidGrantException('"refresh_token" was used before');
            }

            // log the use of the refresh_token so it can't be used again
            $this->storage->logRefreshToken($refreshToken->authKey(), $refreshToken->tokenId());

            // make sure an access_token never outlives its authorization, i.e.
            // the refresh_token expiry
            $accessTokenExpiresAt = $this->dateTime->add($this->accessTokenExpiry);
            if ($accessTokenExpiresAt > $authorizationInfo->expiresAt()) {
                $accessTokenExpiresAt = $authorizationInfo->expiresAt();
            }

            $accessToken = new AccessToken(
                Base64UrlSafe::encodeUnpadded($this->randomBytes()),
                $refreshToken->authKey(),
                $refreshToken->userId(),
                $refreshToken->clientId(),
                $refreshToken->scope(),
                $accessTokenExpiresAt,
                $authorizationInfo->expiresAt()
            );

            $refreshToken = new RefreshToken(
                Base64UrlSafe::encodeUnpadded($this->randomBytes()),
                $refreshToken->authKey(),
                $refreshToken->userId(),
                $refreshToken->clientId(),
                $refreshToken->scope(),
                $authorizationInfo->expiresAt()
            );

            return new JsonResponse(
                [
                    'access_token' => $this->signer->sign($accessToken->json()),
                    'refresh_token' => $this->signer->sign($refreshToken->json()),
                    'token_type' => 'bearer',
                    'expires_in' => $this->expiresAtToInt($accessTokenExpiresAt),
                ],
                // The authorization server MUST include the HTTP "Cache-Control"
                // response header field [RFC2616] with a value of "no-store" in any
                // response containing tokens, credentials, or other sensitive
                // information, as well as the "Pragma" response header field [RFC2616]
                // with a value of "no-cache".
                [
                    'Cache-Control' => 'no-store',
                    'Pragma' => 'no-cache',
                ]
            );
        } catch (SignerException $e) {
            throw new InvalidGrantException(sprintf('invalid "refresh_token": %s', $e->getMessage()));
        }
    }

    private function verifyAuthorizationCode(Request $request, AuthorizationCode $authorizationCode): void
    {
        if ($request->post()->clientId() !== $authorizationCode->clientId()) {
            throw new InvalidRequestException('unexpected "client_id"');
        }

        if ($request->post()->redirectUri() !== $authorizationCode->redirectUri()) {
            throw new InvalidRequestException('unexpected "redirect_uri"');
        }
    }

    private function verifyClientCredentials(string $clientId, ClientInfo $clientInfo, Request $request): void
    {
        $clientSecret = $clientInfo->clientSecret();
        if (null !== $clientSecret) {
            if (null === $authUser = $request->authUser()) {
                throw new InvalidClientException('invalid credentials (no authenticating user)');
            }
            if ($clientId !== $authUser) {
                throw new InvalidClientException('"client_id" does not match authenticating user');
            }
            if (null === $authPass = $request->authPass()) {
                throw new InvalidClientException('invalid credentials (no authenticating pass)');
            }
            if (false === hash_equals($clientSecret, $authPass)) {
                throw new InvalidClientException('invalid credentials (invalid authenticating pass)');
            }
        }
    }

    /**
     * @see https://tools.ietf.org/html/rfc7636#appendix-A
     */
    private function verifyCodeVerifier(Request $request, AuthorizationCode $authorizationCode): void
    {
        // in >= 6.0.0 we ALWAYS require PKCE. In case clients do NOT support
        // PKCE they will fail anyway due to missing authorize/token
        // parameters regarding PKCE and never reach this point. Confidential
        // clients that DO support PKCE will keep working nicely, but in
        // addition now their code verifier is checked as well...
        $strCmp = hash_equals(
            $authorizationCode->codeChallenge(),
            Base64UrlSafe::encodeUnpadded(
                hash(
                    'sha256',
                    $request->post()->codeVerifier(),
                    true
                )
            )
        );

        if (false === $strCmp) {
            throw new InvalidGrantException('invalid "code_verifier"');
        }
    }

    private function getClient(string $clientId): ClientInfo
    {
        if (null === $clientInfo = $this->clientDb->get($clientId)) {
            throw new InvalidClientException('client does not exist with this "client_id"');
        }

        return $clientInfo;
    }

    private function expiresAtToInt(DateTimeImmutable $expiresAt): int
    {
        return (int) ($expiresAt->getTimestamp() - $this->dateTime->getTimestamp());
    }

    /**
     * @param array<string,string> $queryParameters
     */
    private function prepareRedirectUri(string $redirectUri, array $queryParameters): string
    {
        // add the Issuer Identity if set
        if (null !== $this->issuerIdentity) {
            $queryParameters['iss'] = $this->issuerIdentity;
        }

        return sprintf(
            '%s%s%s',
            $redirectUri,
            false === strpos($redirectUri, '?') ? '?' : '&',
            http_build_query($queryParameters)
        );
    }
}
